from shiv.bootstrap import import_string

from site import addsitedir
from code import interact


class TestBootstrap:
    def test_various_imports(self):
        assert import_string('site.addsitedir') == addsitedir
        assert import_string('site:addsitedir') == addsitedir
        assert import_string('code.interact') == interact
        assert import_string('code:interact') == interact

        # test things not already imported
        func = import_string('os.path:join')
        from os.path import join
        assert func == join



import copy
import json
import os

from pathlib import Path


class Environment:
    INTERPRETER = "SHIV_INTERPRETER"
    ENTRY_POINT = "SHIV_ENTRY_POINT"
    ROOT = "SHIV_ROOT"
    FORCE_EXTRACT = "SHIV_FORCE_EXTRACT"

    def __init__(
        self,
        build_id=None,
        entry_point=None,
        shared_object_map=None,
        always_write_cache=False,
    ):
        self.build_id = build_id
        self.shared_object_map = shared_object_map
        self.always_write_cache = always_write_cache

        # properties
        self._entry_point = entry_point

    @classmethod
    def from_json(cls, json_data):
        return Environment(**json.loads(json_data))

    def to_json(self):
        d = copy.copy(self.__dict__)
        del d["_entry_point"]
        d["entry_point"] = self.entry_point
        return json.dumps(d)

    @property
    def entry_point(self):
        return os.environ.get(self.ENTRY_POINT, self._entry_point)

    @property
    def interpreter(self):
        return os.environ.get(self.INTERPRETER, None)

    @property
    def root(self):
        root = os.environ.get(self.ROOT)
        return Path(root) if root is not None else None

    @property
    def force_extract(self):
        return bool(os.environ.get(self.FORCE_EXTRACT, self.always_write_cache))



import site
import sys
import shutil

from compileall import compile_dir
from importlib import import_module
from pathlib import Path

from .environment import Environment
from .interpreter import execute_interpreter
from .utils import current_zipfile


def import_string(import_name):
    """Returns a callable for a given setuptools style import string

    :param import_name: A console_scripts style import string
    """
    import_name = str(import_name).replace(":", ".")
    try:
        import_module(import_name)
    except ImportError:
        if "." not in import_name:
            # this is a case like "import name", where continuing to the
            # next style of import would not improve the situation, so
            # we raise here.
            raise

    else:
        return sys.modules[import_name]

    # this is a case where the previous attempt may have failed due to
    # not being importable. ("not a package", etc)
    module_name, obj_name = import_name.rsplit(".", 1)
    try:
        module = __import__(module_name, None, None, [obj_name])
    except ImportError:
        # Recurse to support importing modules not yet set up by the parent module
        # (or package for that matter)
        module = import_string(module_name)

    try:
        return getattr(module, obj_name)

    except AttributeError as e:
        raise ImportError(e)


def cache_path(archive, root_dir, build_id):
    """Returns a ~/.shiv cache directory for unzipping site-packages during bootstrap.

    :param ZipFile archive: The zipfile object we are bootstrapping from.
    :param str buidl_id: The build id generated at zip creation.
    """
    root = root_dir or Path("~/.shiv").expanduser()
    name = Path(archive.filename).stem
    return root / f"{name}_{build_id}"


def extract_site_packages(archive, target_path):
    """Extract everything in site-packages to a specified path.

    :param ZipFile archive: The zipfile object we are bootstrapping from.
    :param Path target_path: The path to extract our zip to.
    """
    target_path_tmp = Path(target_path.parent, target_path.stem + ".tmp")

    for filename in archive.namelist():
        if filename.startswith("site-packages"):
            archive.extract(filename, target_path_tmp)

    # compile pyc
    compile_dir(target_path_tmp, quiet=1)

    # atomic move
    shutil.move(target_path_tmp.as_posix(), target_path.as_posix())


def bootstrap():
    """Actually bootstrap our shiv environment."""

    # get a handle of the currently executing zip file
    archive = current_zipfile()

    # create an environment object (a combination of env vars and json metadata)
    env = Environment.from_json(archive.read("environment.json").decode())

    # get a site-packages directory (from env var or via build id)
    site_packages = cache_path(archive, env.root, env.build_id) / "site-packages"

    # determine if first run or forcing extract
    if not site_packages.exists() or env.force_extract:
        extract_site_packages(archive, site_packages.parent)

    # stdlib blessed way of extending path
    site.addsitedir(site_packages)

    # do entry point import and call
    if env.entry_point is not None and env.interpreter is None:
        mod = import_string(env.entry_point)
        try:
            mod()
        except TypeError as e:
            # catch "<module> is not callable", which is thrown when the entry point's
            # callable shares a name with it's parent module
            # e.g. "from foo.bar import bar; bar()"
            getattr(mod, env.entry_point.replace(":", ".").split(".")[1])()
    else:
        # drop into interactive mode
        execute_interpreter()


if __name__ == "__main__":
    bootstrap()



"""This module contains a simplified implementation of Python's "zipapp" module.

We've copied code here in order to patch in support for compression.
"""
import contextlib
import zipfile
import stat
import sys
import zipapp

from pathlib import Path
from typing import Any, IO, Generator, Union

# Typical maximum length for a shebang line
BINPRM_BUF_SIZE = 128

# zipapp __main__.py template
MAIN_TEMPLATE = """\
# -*- coding: utf-8 -*-
import {module}
{module}.{fn}()
"""


def write_file_prefix(f: IO[Any], interpreter_path: Path) -> None:
    """Write a shebang line.

    :param f: An open file handle.
    :param interpreter_path: A path to a python interpreter.
    """

    # fall back to /usr/bin/env if the interp path is too long
    if len(interpreter_path.as_posix()) > BINPRM_BUF_SIZE:
        shebang = f"/usr/bin/env {interpreter_path.name}"
    else:
        shebang = interpreter_path.as_posix()

    f.write(b"#!" + shebang.encode(sys.getfilesystemencoding()) + b"\n")


@contextlib.contextmanager
def maybe_open(archive: Union[str, Path], mode: str) -> Generator[IO[Any], None, None]:
    if isinstance(archive, (str, Path)):
        with Path(archive).open(mode=mode) as f:
            yield f

    else:
        yield archive


def create_archive(
    source: Path, target: Path, interpreter: Path, main: str, compressed: bool = True
) -> None:
    """Create an application archive from SOURCE."""

    # Check that main has the right format.
    mod, sep, fn = main.partition(":")
    mod_ok = all(part.isidentifier() for part in mod.split("."))
    fn_ok = all(part.isidentifier() for part in fn.split("."))
    if not (sep == ":" and mod_ok and fn_ok):
        raise zipapp.ZipAppError("Invalid entry point: " + main)

    main_py = MAIN_TEMPLATE.format(module=mod, fn=fn)

    with maybe_open(target, "wb") as fd:
        # write shebang
        write_file_prefix(fd, interpreter)

        # determine compression
        compression = zipfile.ZIP_DEFLATED if compressed else zipfile.ZIP_STORED

        # create zipapp
        with zipfile.ZipFile(fd, "w", compression=compression) as z:
            for child in source.rglob("*"):

                # skip compiled files
                if child.suffix == '.pyc':
                    continue

                arcname = child.relative_to(source)
                z.write(child.as_posix(), arcname.as_posix())

            # write main
            z.writestr("__main__.py", main_py.encode("utf-8"))

    # make executable
    if interpreter and not hasattr(target, "write"):
        target.chmod(target.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)



import tempfile
import stat
import sys
import zipfile

from pathlib import Path
from zipapp import ZipAppError

import pytest

from shiv.cli import validate_interpreter
from shiv.builder import write_file_prefix, create_archive


UGOX = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH


class TestBuilder:
    def test_file_prefix(self):
        with tempfile.TemporaryFile() as fd:
            python = validate_interpreter(Path(sys.executable))
            write_file_prefix(fd, python)
            fd.seek(0)
            written = fd.read()

        assert written == b'#!' + python.as_posix().encode(sys.getdefaultencoding()) + b'\n'

    def test_create_archive(self, sp):
        with tempfile.TemporaryDirectory() as tmpdir:
            target = Path(tmpdir, 'test.zip')
            create_archive(sp, target, validate_interpreter(None), 'code:interact')

            assert zipfile.is_zipfile(str(target))

            with pytest.raises(ZipAppError):
                create_archive(sp, target, validate_interpreter(None), 'alsjdbas,,,')

    def test_archive_permissions(self, sp):
        with tempfile.TemporaryDirectory() as tmpdir:
            target = Path(tmpdir, 'test.zip')
            create_archive(sp, target, validate_interpreter(None), 'code:interact')

            assert target.stat().st_mode & UGOX == UGOX



import importlib_resources  # type: ignore
import os
import shutil
import sys
import uuid

from configparser import ConfigParser
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Optional, List

import click

from . import pip
from . import builder
from . import bootstrap
from .bootstrap.environment import Environment
from .constants import (
    BLACKLISTED_ARGS,
    DISALLOWED_PIP_ARGS,
    NO_PIP_ARGS,
    NO_OUTFILE,
    NO_ENTRY_POINT,
    INVALID_PYTHON,
)

# This is the 'knife' emoji
SHIV = u"\U0001F52A"


def find_entry_point(site_packages: Path, console_script: str) -> str:
    """Find a console_script in a site-packages directory.

    Console script metadata is stored in entry_points.txt per setuptools
    convention. This function searches all entry_points.txt files and
    returns the import string for a given console_script argument.

    :param site_packages: A path to a site-packages directory on disk.
    :param console_script: A console_script string.
    """
    config_parser = ConfigParser()
    config_parser.read(site_packages.rglob("entry_points.txt"))
    return config_parser["console_scripts"][console_script]


def validate_interpreter(interpreter_path: Optional[str] = None) -> Path:
    """Ensure that the interpreter is a real path, not a symlink.

    If no interpreter is given, default to `sys.exectuable`

    :param interpreter_path: A path to a Python interpreter.
    """
    real_path = Path(sys.executable) if interpreter_path is None else Path(
        interpreter_path
    )

    if real_path.exists():
        return real_path

    else:
        sys.exit(INVALID_PYTHON.format(path=real_path))


def map_shared_objects(site_packages: Path) -> Dict[str, str]:
    """Given a site-packages dir, map all of the shared objects to their namespaces.

    :param site_packages: A path to a custom site-packages directory.
    """
    somap: Dict[str, str] = {}

    for parent_dir, _, filenames in os.walk(site_packages):

        for filename in filenames:

            # get full path to file
            fullpath = Path(parent_dir) / filename

            # check if the file is a shared object (skipping if not)
            if fullpath.suffix not in [".so", ".dylib"]:
                continue

            # assemble the mapping of import path to file
            index = fullpath.parts.index("site-packages") + 1
            contributors = list(fullpath.parts[index:-1])
            module_name, import_tag, extension = fullpath.parts[-1].split(".")
            contributors.append(module_name)
            import_path = ".".join(contributors)

            # finally, add the import path to the shared object map
            somap[import_path] = fullpath.relative_to(site_packages).as_posix()

    return somap


def copy_bootstrap(bootstrap_target: Path) -> None:
    """Copy bootstrap code from shiv into the pyz.

    :param bootstrap_target: The temporary directory where we are staging pyz contents.
    """
    for bootstrap_file in importlib_resources.contents(bootstrap):
        if importlib_resources.is_resource(bootstrap, bootstrap_file):
            with importlib_resources.path(bootstrap, bootstrap_file) as f:
                shutil.copyfile(f.absolute(), bootstrap_target / f.name)


@click.command(
    context_settings=dict(
        help_option_names=["-h", "--help", "--halp"], ignore_unknown_options=True
    )
)
@click.option("--entry-point", "-e", default=None, help="The entry point to invoke.")
@click.option(
    "--console-script", "-c", default=None, help="The console_script to invoke."
)
@click.option("--output-file", "-o", help="The file for shiv to create.")
@click.option("--python", "-p", help="The path to a python interpreter to use.")
@click.option(
    "--compressed/--uncompressed",
    default=True,
    help="Whether or not to compress your zip.",
)
@click.argument("pip_args", nargs=-1, type=click.UNPROCESSED)
def main(
    output_file: str,
    entry_point: Optional[str],
    console_script: Optional[str],
    python: Optional[str],
    compressed: bool,
    pip_args: List[str],
) -> None:
    """
    Shiv is a command line utility for building fully self-contained Python zipapps
    as outlined in PEP 441, but with all their dependencies included!
    """
    quiet = "-q" in pip_args

    if not quiet:
        click.secho(" shiv! " + SHIV, bold=True)

    if not pip_args:
        sys.exit(NO_PIP_ARGS)

    if output_file is None:
        sys.exit(NO_OUTFILE)

    # check for disallowed pip arguments
    for blacklisted_arg in BLACKLISTED_ARGS:
        for supplied_arg in pip_args:
            if supplied_arg in blacklisted_arg:
                sys.exit(
                    DISALLOWED_PIP_ARGS.format(
                        arg=supplied_arg, reason=BLACKLISTED_ARGS[blacklisted_arg]
                    )
                )

    # validate supplied python (if any)
    interpreter = validate_interpreter(python)

    with TemporaryDirectory() as working_path:
        site_packages = Path(working_path, "site-packages")
        site_packages.mkdir(parents=True, exist_ok=True)

        # install deps into staged site-packages
        pip.install(
            python or sys.executable,
            ["--target", site_packages.as_posix()] + list(pip_args),
        )

        # if entry_point is a console script, get the callable
        if entry_point is None and console_script is not None:
            try:
                entry_point = find_entry_point(site_packages, console_script)
            except KeyError:
                sys.exit(NO_ENTRY_POINT.format(entry_point=console_script))

        # create runtime environment metadata
        env = Environment(
            build_id=str(uuid.uuid4()),
            entry_point=entry_point,
            shared_object_map=map_shared_objects(site_packages),
        )

        Path(working_path, "environment.json").write_text(env.to_json())

        # create bootstrapping directory in working path
        bootstrap_target = Path(working_path, "_bootstrap")
        bootstrap_target.mkdir(parents=True, exist_ok=True)

        # copy bootstrap code
        copy_bootstrap(bootstrap_target)

        # create the zip
        builder.create_archive(
            Path(working_path),
            target=Path(output_file),
            interpreter=interpreter,
            main="_bootstrap:bootstrap",
            compressed=compressed,
        )

    if not quiet:
        click.secho(" done ", bold=True)



import sys
import zipfile


def current_zipfile():
    """A function to vend the current zipfile, if any"""
    if zipfile.is_zipfile(sys.argv[0]):
        fd = open(sys.argv[0], "rb")
        return zipfile.ZipFile(fd)


def loaded_from_zipfile(module):
    return hasattr(module.__spec__.loader, "archive")



import os
import re
import setuptools
import sys
import venv

from pathlib import Path
from setuptools.command import easy_install

if sys.version_info < (3, 6):
    print('shiv requires at least Python 3.6!')
    sys.exit(1)

requirements = [
    'click==6.7',
    'pip>=9.0.1',
    'importlib_resources>=0.4',
]

# The following template and classmethod are copied from
# fast entry points, Copyright (c) 2016, Aaron Christianson
# https://github.com/ninjaaron/fast-entry_point

TEMPLATE = '''\
# -*- coding: utf-8 -*-
# EASY-INSTALL-ENTRY-SCRIPT: '{3}','{4}','{5}'
__requires__ = '{3}'
import re
import sys
from {0} import {1}
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit({2}())'''


@classmethod
def get_args(cls, dist, header=None):
    """Overrides easy_install.ScriptWriter.get_args

    This method avoids using pkg_resources to map a named entry_point
    to a callable at invocation time.
    """
    if header is None:
        header = cls.get_header()
    spec = str(dist.as_requirement())
    for type_ in 'console', 'gui':
        group = type_ + '_scripts'
        for name, ep in dist.get_entry_map(group).items():
            # ensure_safe_name
            if re.search(r'[\\/]', name):
                raise ValueError("Path separators not allowed in script names")
            script_text = TEMPLATE.format(
                ep.module_name,
                ep.attrs[0],
                '.'.join(ep.attrs),
                spec,
                group,
                name,
            )
            args = cls._get_script_args(type_, name, header, script_text)
            for res in args:
                yield res

# patch in the fast-entry_points classmethod
easy_install.ScriptWriter.get_args = get_args

# fast entry points end.


class Venv(setuptools.Command):
    user_options = []

    def initialize_options(self):
        """Abstract method that is required to be overwritten"""

    def finalize_options(self):
        """Abstract method that is required to be overwritten"""

    def run(self):
        venv_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'venv', 'shiv')
        print('Creating virtual environment in {}'.format(venv_path))
        venv.main(args=[venv_path])
        print(
            'Linking `activate` to top level of project.\n'
            'To activate, simply run `source activate`.'
        )
        try:
            os.symlink(
                Path(venv_path, 'bin', 'activate'),
                Path(Path(__file__).absolute().parent, 'activate'),
            )
        except FileExistsError:
            ...


setuptools.setup(
    name='shiv',
    version='0.0.8',
    description="A command line utility for building fully self contained Python zipapps.",
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'shiv = shiv.cli:main',
        ],
    },
    include_package_data=True,
    cmdclass={'venv': Venv},
    license='BSD License',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)
